#!/usr/bin/env bash
set -Eeuo pipefail
trap 'echo "ERROR: $BASH_SOURCE:$LINENO $BASH_COMMAND" >&2' ERR

source "/usr/lib/hades/functions.sh"

print_usage() {
	msg "\
Usage: $0 [-h] [--help] { start | stop | reload | clear }

Control the hades database PostgreSQL cluster.

Commands:
  init                  Initialize the database
  start                 Start the database
  stop                  Stop the database
  reload                Reload the database
  clear                 Clear the database

Options
  -h --help             Print this message
  --clear               Delete existing cluster before"
}

clear_pgdata() {
	msg "Removing PostgreSQL cluster at $PGDATA"
	shopt -s dotglob
	/bin/rm -rf "$PGDATA"/*
	shopt -u dotglob
}

cleanup() {
	error "Aborting database initialisation"
	/usr/lib/postgresql/9.6/bin/pg_ctl stop -s || true
	clear_pgdata
}

cluster_exists() {
	[[ -f "$PGDATA/PG_VERSION" && -d "$PGDATA/base" ]]
}

check_cluster_version() {
	local CLUSTER_VERSION=
	read -r CLUSTER_VERSION < "$PGDATA/PG_VERSION" || :
	local PG_VERSION="$(/usr/lib/postgresql/9.6/bin/postgres --version | /bin/sed 's/.* \([0-9]\+\.[0-9]\+\).[0-9]\+/\1/')"
	if [[ "$PG_VERSION" != "$CLUSTER_VERSION" ]]; then
		error "Error: version of PostgreSQL cluster located at $PGDATA ($CLUSTER_VERSION) does not match version of postgres executable ($PG_VERSION)"
	fi
}

generate_config() {
	/usr/bin/hades-generate-config postgresql "/run/hades/database/conf"
}

do_init() {
	if cluster_exists; then
		check_cluster_version
		msg "Skipping database initialisation. PostgreSQL cluster already exists."
		return "$EX_OK"
	fi

	trap cleanup EXIT HUP INT QUIT ABRT
	msg "Initialising new PostgreSQL cluster at $PGDATA"
	TZ=Etc/UTC /usr/lib/postgresql/9.6/bin/pg_ctl initdb -s -o "--auth-host=password --auth-local=peer --encoding=UTF-8 --locale=C"
	/bin/rm "$PGDATA"/postgresql.conf "$PGDATA"/pg_hba.conf "$PGDATA"/pg_ident.conf
	generate_config
	/usr/lib/postgresql/9.6/bin/pg_ctl start -w -s -o '-c config_file=/run/hades/database/conf/postgresql.conf'
	for user in 'hades-agent' 'hades-portal' 'hades-radius'; do
		/usr/lib/postgresql/9.6/bin/createuser "$user"
	done

	if [[ "$HADES_LOCAL_MASTER_DATABASE" = True ]]; then
		msg "Creating local master database foreign for testing."
		/usr/lib/postgresql/9.6/bin/createdb foreign
		/usr/bin/psql --quiet --set=ON_ERROR_STOP=1 --no-psqlrc --single-transaction postgres <<-COMMAND
			CREATE ROLE "foreign" WITH LOGIN INHERIT UNENCRYPTED PASSWORD 'foreign';
			COMMAND
		msg "Loading local master database schema"
		/usr/bin/hades-generate-config schema_fdw.sql.j2 | /usr/bin/psql --quiet --set=ON_ERROR_STOP=1 --no-psqlrc --single-transaction --file=- foreign
	fi

	msg "Creating database hades"
	/usr/lib/postgresql/9.6/bin/createdb hades
	msg "Loading database schema"
	/usr/bin/hades-generate-config schema.sql.j2 | /usr/bin/psql --quiet --set=ON_ERROR_STOP=1 --no-psqlrc --single-transaction --file=- hades
	/usr/lib/postgresql/9.6/bin/pg_ctl stop -s
	msg "PostgreSQL cluster at $PGDATA successfully initialised."
	trap - EXIT HUP INT QUIT ABRT
	return "$EX_OK"
}

do_start() {
	if /usr/lib/postgresql/9.6/bin/pg_ctl status &>/dev/null; then
		error "Error: PostgreSQL cluster already running"
		return 3
	fi
	if cluster_exists; then
		check_cluster_version
	else
		error "Error: PostgreSQL cluster at $PGDATA does not exists"
		return 2
	fi
	generate_config
	/usr/lib/postgresql/9.6/bin/pg_ctl start -s -w -t 30 -o '-c config_file=/run/hades/database/conf/postgresql.conf' ${POSTGRESQL_OPTS:+"-o ${POSTGRESQL_OPTS}"}
}

do_stop() {
	/usr/lib/postgresql/9.6/bin/pg_ctl stop -s -t 30 -m fast
}

do_reload() {
	generate_config
	/usr/lib/postgresql/9.6/bin/pg_ctl reload -s
}

do_clear() {
	if /usr/lib/postgresql/9.6/bin/pg_ctl status &>/dev/null; then
		error "Error: can't clear running PostgreSQL cluster"
		return 3
	fi
	clear_pgdata
}

main() {
	if (( $# != 1)); then
		print_usage
		exit "$EX_USAGE"
	fi
	case "$1" in
		-h|--help|help)
			print_usage
			exit "$EX_OK"
			;;
		init|start|stop|reload|clear)
			load_config
			export PGHOST="/run/hades/database"
			export PGPORT="$HADES_POSTGRESQL_PORT"
			export PGDATA="/var/lib/hades/database"
			do_$1 "$@"
			;;
		*)
			print_usage
			exit "$EX_USAGE"
			;;
	esac
}

main "$@"
