#!/usr/bin/env bash
set -Eeuo pipefail
trap 'echo "ERROR: $BASH_SOURCE:$LINENO $BASH_COMMAND" >&2' ERR

source "/usr/lib/hades/functions.sh"

print_usage() {
	msg "\
Usage: $0 { start | stop }

Commands:
  start                 Configure hades network settings
  stop                  Deconfigure hades network settings

Options:
  -h --help             Print this message"
}

add_address() {
	local -r namespace="$1"
	local -r interface="$2"
	local -r address="$3"
	local -r peer_address="${4-}"

	local -a netns=()
	if [[ -z "$namespace" ]]; then
		msg "Adding $address to $interface"
	else
		msg "Adding $address to $interface in namespace $namespace"
		netns=(/sbin/ip netns exec "$namespace")
	fi
	${netns+"${netns[@]-}"} /sbin/ip address add "$address" ${peer_address:+peer "$peer_address"} dev "$interface"
}

netaddr.ip() {
	/usr/bin/python3 -c "import sys, netaddr; print(netaddr.IPNetwork(sys.argv[1]).ip)" "$1"
}

netaddr.cidr() {
	/usr/bin/python3 -c "import sys, netaddr; print(netaddr.IPNetwork(sys.argv[1]).cidr)" "$1"
}

create_bridge() {
	local -r interface="$1"
	local -r address="${2-}"
	local -r sysfs_path="/sys/class/net/$interface"
	msg "Creating bridge $interface"
	/sbin/ip link add "$interface" type bridge
	# Like brctl setfd
	echo 0 > "$sysfs_path/bridge/forward_delay"
	# Like brctl stp
	echo 0 > "$sysfs_path/bridge/stp_state"
	if [[ -n "$address" ]]; then
		add_address "" "$interface" "$address"
	fi
	/sbin/ip link set up dev "$interface"
}

attach_bridge_port() {
	local -r bridge="$1"
	local -r slave="$2"
	msg "Attaching interface $slave to bridge $bridge"
	/sbin/ip link set "$slave" master "$bridge"
}

link_namespace() {
	local -r namespace="$1"
	local -r parent_interface="$2"
	local -r namespace_interface="$3"
	local -r parent_address="${4-}"
	local -r namespace_address="${5-}"

	msg "Linking network namespace $namespace via veth pair $parent_interface <-> $namespace_interface"
	/sbin/ip link add name "$parent_interface" type veth peer netns "$namespace" name "$namespace_interface"
	if [[ -n "$parent_address" ]]; then
		add_address "" "$parent_interface" "$(netaddr.ip "$parent_address")" ${namespace_address+"$(netaddr.ip "$namespace_address")"}
	fi
	if [[ -n "$namespace_address" ]]; then
		add_address "$namespace" "$namespace_interface" "$(netaddr.ip "$namespace_address")" ${parent_address+"$(netaddr.ip "$parent_address")"}
	fi
	/sbin/ip link set dev "$parent_interface" up
	/sbin/ip netns exec "$namespace" /sbin/ip link set dev "$namespace_interface" up
}

setup_namespace_common_pre() {
	local -r namespace="$1"
	local -r interface="$2"
	local -r parent_address="$3"
	local -r namespace_address="$4"

	msg "Creating network namespace $namespace"
	/sbin/ip netns add "$namespace"
	/sbin/ip netns exec "$namespace" /sbin/ip link set dev lo up
	/sbin/ip netns exec "$namespace" /sbin/sysctl --quiet net.ipv4.ip_nonlocal_bind=1
	/sbin/ip netns exec "$namespace" /sbin/sysctl --quiet net.ipv4.ip_forward=1
	/sbin/ip netns exec "$namespace" /sbin/sysctl --quiet net.ipv6.conf.all.disable_ipv6=1
	/sbin/ip netns exec "$namespace" /sbin/sysctl --quiet net.ipv6.conf.default.disable_ipv6=1

	link_namespace "$namespace" "${namespace}-veth0" eth0 "$parent_address" "$namespace_address"
	/sbin/ip netns exec "$namespace" /sbin/ip route add default via "$(netaddr.ip "$parent_address")"

	link_namespace "$namespace" "${namespace}-veth1" eth1
	attach_bridge_port "$HADES_VRRP_BRIDGE" "${namespace}-veth1"

	create_bridge "br-${namespace}"
	attach_bridge_port "br-${namespace}" "$interface"
	link_namespace "$namespace" "${namespace}-veth2" eth2
	attach_bridge_port "br-${namespace}" "${namespace}-veth2"
}

setup_namespace_auth() {
	msg "Creating ipset"
	/sbin/ip netns exec auth /sbin/ipset create "$HADES_AUTH_DNS_ALTERNATIVE_IPSET" hash:ip
}

setup_namespace_unauth() {
	msg "Creating ipset"
	/sbin/ip netns exec unauth /sbin/ipset create "$HADES_UNAUTH_WHITELIST_IPSET" hash:ip
	/sbin/sysctl -w net.ipv4.conf.br-unauth.log_martians=0
}

setup_namespace_common_post() {
	local -r namespace="$1"
	local -r interface="$2"

	msg "Loading iptables rules for network namespace $namespace"
	/usr/bin/hades-generate-config "iptables-${namespace}.j2" | /sbin/ip netns exec "$namespace" /sbin/iptables-restore
}

setup_host_network() {
	/sbin/sysctl --quiet net.ipv4.ip_nonlocal_bind=1
	/sbin/sysctl --quiet net.ipv4.ip_forward=1
	if [[ "$HADES_CREATE_DUMMY_INTERFACES" = True ]]; then
		local interface
		for interface in "$HADES_RADIUS_INTERFACE" "$HADES_VRRP_INTERFACE" "$HADES_AUTH_INTERFACE" "$HADES_UNAUTH_INTERFACE"; do
			if [[ -d "/sys/class/net/$interface" ]]; then
				continue
			fi
			msg "Creating dummy interface $interface"
			/sbin/ip link add name "$interface" type dummy
			/sbin/ip link set up dev "$interface"
		done
	fi
	create_bridge "$HADES_VRRP_BRIDGE" "$HADES_VRRP_LISTEN_RADIUS"
	attach_bridge_port "$HADES_VRRP_BRIDGE" "$HADES_VRRP_INTERFACE"

	/usr/bin/hades-generate-config iptables-main.j2 | /sbin/iptables-restore
}

do_start() {
	setup_host_network
	setup_namespace_common_pre auth   "$HADES_AUTH_INTERFACE"   "$HADES_NETNS_MAIN_AUTH_LISTEN"   "$HADES_NETNS_AUTH_LISTEN"
	setup_namespace_common_pre unauth "$HADES_UNAUTH_INTERFACE" "$HADES_NETNS_MAIN_UNAUTH_LISTEN" "$HADES_NETNS_UNAUTH_LISTEN"
	setup_namespace_auth
	setup_namespace_unauth
	setup_namespace_common_post auth   "$HADES_AUTH_INTERFACE"
	setup_namespace_common_post unauth "$HADES_UNAUTH_INTERFACE"
}

teardown_namespace() {
	local -r namespace="$1"
	local -r interface="$2"
	local -a pids=()
	if ! /sbin/ip netns pids "$namespace" | readarray -t pids; then
		return
	fi
	if (( ${#pids[@]} > 0 )); then
		error "Error: Network namespace $namespace has still processes running. PIDs: ${pids[*]}"
		return 2
	fi

	msg "Deleting network namespace $namespace"
	/sbin/ip link delete dev "${namespace}-veth0"
	/sbin/ip link delete dev "${namespace}-veth1"
	/sbin/ip link delete dev "${namespace}-veth2"
	/sbin/ip link delete dev "br-${namespace}"
	/sbin/ip netns delete "$namespace"
}

do_stop() {
	teardown_namespace auth   "$HADES_AUTH_INTERFACE"
	teardown_namespace unauth "$HADES_UNAUTH_INTERFACE"
	/sbin/ip link delete dev "$HADES_VRRP_BRIDGE"

	msg "Resetting iptables"
	local -a tables=()
	readarray -t tables < /proc/net/ip_tables_names
	local -Ar chains=(
		['raw']='PREROUTING OUTPUT'
		['mangle']='PREROUTING INPUT FORWARD OUTPUT POSTROUTING'
		['nat']='PREROUTING INPUT OUTPUT POSTROUTING'
		['filter']='INPUT FORWARD OUTPUT'
		['security']='INPUT FORWARD OUTPUT'
	)
	for table in "${tables[@]}"; do
		/sbin/iptables --table "$table" --flush
		/sbin/iptables --table "$table" --delete-chain
		for chain in ${chains[$table]}; do
			/sbin/iptables --table "$table" --policy "$chain" ACCEPT
		done
	done
}

main() {
	if (( $# != 1)); then
		print_usage
		exit "$EX_USAGE"
	fi
	case "$1" in
		-h|--help)
			print_usage
			exit "$EX_OK"
			;;
		start|stop)
			load_config
			do_$1
			;;
		*)
			print_usage
			exit "$EX_USAGE"
			;;
	esac
}

main "$@"
